import os, time, json, shutil, traceback
from datetime import datetime
from random import randint
from werkzeug.security import generate_password_hash, check_password_hash
from flask import request, abort, redirect, Response, url_for, render_template
from flask_login import UserMixin, login_required, UserMixin, current_user
from app import login_manager
from .error import logger, getHTMLErrorString
from .roles import ROLES

# html generator needs roles to be initied for it to run
# Otherwise it's a race condition and will result in import error
from .html_generator import make_top_bar

USERS_DIR = "content/users"
os.makedirs(USERS_DIR, exist_ok=True)


class adminToken:
    def __init__(self, token_string, lifetime=60 * 15):
        self.token = token_string
        self.expiration = time.time() + lifetime

    def get(self, value):
        if self.expiration <= time.time():
            return False
        return True if self.token == value else False

    def check_validity(self):
        if self.expiration >= time.time():
            return True


class singleTokenGenerator:
    def __init__(self):
        self.token = None

    # Invalidates any previous token
    def generate_token(self):
        self.token = adminToken(str(randint(10000, 99999)))
        return self.token

    def check_token(self, token_string):
        if self.token:
            return self.token.get(token_string)

    def get_token(self):
        if self.token:
            if self.token.check_validity():
                return self.token
        self.token = None

    def invalidate_token(self):
        self.token = None


token_generator = singleTokenGenerator()


class baseUser(UserMixin):
    def __init__(
        self,
        username="INVALID",
        password_hash="0000000000",
        account_banned=False,
        active=False,
        id=None,
        role=ROLES.OWNER.value,  # Role in server
        created=None,
        created_friendly=None,
        bio=None,
        flair=None,
    ):
        self.username = username
        self.password_hash = password_hash
        self.active = active
        self.id = id
        self.account_banned = account_banned
        self.role = int(role)
        self.created = created
        self.created_friendly = created_friendly
        self.bio = bio
        self.flair = flair

    def get_id(self):
        return self.id

    def is_active(self):
        return self.active

    def get_auth_token(self):
        return make_secure_token(self.username, key="secret_key")

    def set_password(self, password):
        self.password_hash = generate_password_hash(password)

    def check_password(self, password):
        return check_password_hash(self.password_hash, password)

    def json(self):
        return {
            "id": self.id,
            "username": self.username,
            "created": self.created,
            "created_friendly": datetime.strftime(
                datetime.utcfromtimestamp(int(float(self.created))), "%m-%d-%y %H:%M:%S"
            ),
            "password_hash": self.password_hash,
            "role": str(self.role),
            "account_banned": self.account_banned,
            "active": self.active,
            "bio": self.bio,
            "flair": self.flair,
        }

    def save(self):
        userdir = os.path.join(USERS_DIR, self.id)
        os.makedirs(userdir, exist_ok=True)
        with open(os.path.join(userdir, "user.json"), "w+") as f:
            json.dump(self.json(), f)


class newUser(baseUser):
    def __init__(self, username, password_hash, active=True):
        baseUser.__init__(
            self,
            username,
            password_hash,
            active=active,
            created=str(time.time()),
            id=str(randint(100000000000, 999999999999)),
            role=ROLES.USER.value,
        )
        self.username = username
        self.active = active
        self.save()
        with open(os.path.join(USERS_DIR, self.id, "posts.json"), "w+") as f:
            json.dump({}, f)


class newOwner(newUser):
    def __init__(self, *args, **kwargs):
        newUser.__init__(self, *args, **kwargs)
        self.role = ROLES.OWNER.value
        self.save()


# Load returning user from file
class returningUser(baseUser):
    def __init__(self, id):
        with open(os.path.join(USERS_DIR, id, "user.json"), "r") as f:
            baseUser.__init__(self, *[], **json.load(f))


class UsersRepository:
    def __init__(self):
        self.user_ids = []
        self.scan()

    def scan(self):
        self.user_ids = [p.name for p in os.scandir(USERS_DIR)]

    def get_user(self, id):
        if id in self.user_ids:
            return returningUser(id)

    def get_user_by_username(self, username):
        for i in self.user_ids:
            u = returningUser(i)
            if u.username.lower() == username.lower():
                return u

    def new_user(self, *args, **kw):
        user = newUser(*args, **kw)
        self.scan()
        print(self.user_ids)
        return user

    def new_owner(self, *args, **kw):
        user = newOwner(*args, **kw)
        self.scan()
        print(self.user_ids)
        return user

    def remove_user(self, id):
        try:
            shutil.rmtree(os.path.join(USERS_DIR, id))
            self.scan()
        except Exception as e:
            logger.log(f"Error removing user {id} - {e}")
            return e

    def set_role(self, id, role):
        try:
            u = self.get_user(id)
            u.role = role
            u.save()
        except Exception as e:
            return e

    def set_bio(self, id, bio):
        try:
            u = self.get_user(id)
            u.bio = bio
            u.save()
        except Exception as e:
            return e

    def set_flair(self, id, flair):
        try:
            u = self.get_user(id)
            u.flair = flair
            u.save()
        except Exception as e:
            return e

    # Returns excsption string on error
    def set_role_banned(self, id):
        return self.set_role(id, ROLES.BANNED.value)

    def set_role_user(self, id):
        return self.set_role(id, ROLES.USER.value)

    def set_role_developer(self, id):
        return self.set_role(id, ROLES.DEVELOPER.value)

    def set_role_moderator(self, id):
        return self.set_role(id, ROLES.MODERATOR.value)

    def set_role_admin(self, id):
        return self.set_role(id, ROLES.ADMIN.value)

    # Returns true if new username is ok
    def check_username_unused(self, username):
        for u in self.user_ids:
            if returningUser(u).username.lower() == username.lower():
                return False
        return True

    def json(self):
        return [returningUser(u).json() for u in self.user_ids]


def owner_only(method):
    def apply(*args, **kw):
        mname = method.__dict__.get("mname") or method.__name__
        try:
            if int(current_user.role) == ROLES.OWNER.value:
                logger.log(
                    f"Verified owner {current_user.username} - {current_user.id} - {mname}"
                )
                return method(*args, **kw)
            logger.log(
                f"Unauthorized user {current_user.username} attempted to access {mname}"
            )
            return render_template(
                "user/unauthorized.html",
                title="Unauthorized",
                subtitle="You're not supposed to be here.",
                error="You do not have permission to view this page.",
                navbar=make_top_bar(current_user),
            )
        except Exception as e:
            title = f"Error running owner check on method {mname} with args {args} and kwargs {kw}"
            string = traceback.format_exc()
            logger.log(title + "\n" + string)
            error = getHTMLErrorString(title, string.replace("\n", "<br/>"))
            return render_template("user/error.html", error=error)

    apply.__dict__["mname"] = method.__dict__.get("mname") or method.__name__
    return apply


def admin_or_above(method):
    def apply(*args, **kw):
        mname = method.__dict__.get("mname") or method.__name__
        try:
            if int(current_user.role) >= ROLES.ADMIN.value:
                logger.log(
                    f"Verified admin {current_user.username} - {current_user.id} - {mname}"
                )
                return method(*args, **kw)
            logger.log(
                f"Unauthorized user {current_user.username} attempted to access {mname}"
            )
            return render_template(
                "user/unauthorized.html",
                title="Unauthorized",
                subtitle="You're not supposed to be here.",
                error="You do not have permission to view this page.",
                navbar=make_top_bar(current_user),
            )
        except Exception as e:
            title = f"Error running admin check on method {mname} with args {args} and kwargs {kw}"
            string = traceback.format_exc()
            logger.log(title + "\n" + string)
            error = getHTMLErrorString(title, string.replace("\n", "<br/>"))
            return render_template(
                "user/error.html",
                title="Check Error",
                subtitle="Issue checking user permissions.",
                error=error,
                navbar=make_top_bar(current_user),
            )

    apply.__dict__["mname"] = method.__dict__.get("mname") or method.__name__
    return apply


def moderator_or_above(method):
    def apply(*args, **kw):
        mname = method.__dict__.get("mname") or method.__name__
        try:
            if int(current_user.role) >= ROLES.MODERATOR.value:
                logger.log(
                    f"Verified moderator {current_user.username} - {current_user.id} - {mname}"
                )
                return method(*args, **kw)
            logger.log(
                f"Unauthorized user {current_user.username} attempted to access {mname}"
            )
            return render_template(
                "user/unauthorized.html",
                title="Unauthorized",
                subtitle="You're not supposed to be here.",
                error="You do not have permission to view this page.",
                navbar=make_top_bar(current_user),
            )
        except Exception as e:
            title = f"Error running moderator check on method {mname} with args {args} and kwargs {kw}"
            string = traceback.format_exc()
            logger.log(title + "\n" + string)
            error = getHTMLErrorString(title, string.replace("\n", "<br/>"))
            return render_template(
                "user/error.html",
                title="Check Error",
                subtitle="Issue checking user permissions.",
                error=error,
                navbar=make_top_bar(current_user),
            )

    apply.__dict__["mname"] = method.__dict__.get("mname") or method.__name__
    return apply


def dev_or_above(method):
    def apply(*args, **kw):
        mname = method.__dict__.get("mname") or method.__name__
        try:
            if int(current_user.role) >= ROLES.DEVELOPER.value:
                logger.log(
                    f"Verified dev {current_user.username} - {current_user.id} - {mname}"
                )
                return method(*args, **kw)
            logger.log(
                f"Unauthorized user {current_user.username} attempted to access {mname}"
            )
            return render_template(
                "user/unauthorized.html",
                title="Unauthorized",
                subtitle="You're not supposed to be here.",
                error="You do not have permission to view this page.",
                navbar=make_top_bar(current_user),
            )
        except Exception as e:
            title = f"Error running dev check on method {mname} with args {args} and kwargs {kw}"
            string = traceback.format_exc()
            logger.log(title + "\n" + string)
            error = getHTMLErrorString(title, string.replace("\n", "<br/>"))
            return render_template(
                "user/error.html",
                error=error,
                title="Check Error",
                subtitle="Issue checking user permissions.",
                navbar=make_top_bar(current_user),
            )

    apply.__dict__["mname"] = method.__dict__.get("mname") or method.__name__
    return apply


def user_or_above(method):
    def apply(*args, **kw):
        mname = method.__dict__.get("mname") or method.__name__
        try:
            if int(current_user.role) >= ROLES.USER.value:
                logger.log(f"Banned user attempted to access {mname}")
                return render_template(
                    "user/unauthorized.html",
                    title="You've been naughty.",
                    subtitle="Santa is mad.",
                    error="You do not have permission to view this page due to being banned.",
                    navbar=make_top_bar(current_user),
                )
            logger.log(
                f"Verified unbanned status {current_user.username} - {current_user.id} - {mname}"
            )
            return method(*args, **kw)
        except Exception as e:
            title = f"Error running ban check on method {mname} with args {args} and kwargs {kw}"
            string = traceback.format_exc()
            logger.log(title + "\n" + string)
            error = getHTMLErrorString(title, string.replace("\n", "<br/>"))
            return render_template(
                "user/error.html",
                title=title,
                subtitle="Ban check error",
                error=error,
                navbar=make_top_bar(current_user),
            )

    apply.__dict__["mname"] = method.__dict__.get("mname") or method.__name__
    return apply


def require_minimum_role_integer(method, role_int):
    def apply(*args, **kw):
        mname = method.__dict__.get("mname") or method.__name__
        try:
            if int(current_user.role) >= role_int:
                logger.log(f"Unauthorized user attempted to access {mname}")
                return render_template(
                    "user/unauthorized.html",
                    title="You've been naughty.",
                    subtitle="Santa is mad.",
                    error="You do not have permission to view this page.",
                    navbar=make_top_bar(current_user),
                )
            logger.log(
                f"Verified access status {current_user.username} - {current_user.id} - {mname}"
            )
            return method(*args, **kw)
        except Exception as e:
            title = f"Error running access check on method {mname} with args {args} and kwargs {kw}"
            string = traceback.format_exc()
            logger.log(title + "\n" + string)
            error = getHTMLErrorString(title, string.replace("\n", "<br/>"))
            return render_template(
                "user/error.html",
                title=title,
                subtitle="Access check error",
                error=error,
                navbar=make_top_bar(current_user),
            )

    apply.__dict__["mname"] = method.__dict__.get("mname") or method.__name__
    return apply


users_repository = UsersRepository()
